import json
import logging
from datetime import datetime
from itertools import groupby
from typing import List, Optional, Union

from openpyxl import load_workbook

from opennem.core.dispatch_type import parse_dispatch_type
from opennem.core.facility.fueltechs import parse_facility_fueltech
from opennem.core.facilitystatus import (
    map_aemo_facility_status,
    parse_facility_status,
)
from opennem.core.fueltechs import lookup_fueltech
from opennem.core.loader import PROJECT_DATA_PATH, load_data
from opennem.core.normalizers import (
    clean_capacity,
    normalize_aemo_region,
    normalize_duid,
    station_name_cleaner,
)
from opennem.core.stations.station_code_from_duids import (
    station_code_from_duids,
)
from opennem.core.stations.station_name_code_map import station_name_code_map
from opennem.core.unit_codes import get_basecode, get_unit_code
from opennem.core.unit_parser import parse_unit_duid
from opennem.exporter.encoders import OpenNEMJSONEncoder
from opennem.schema.stations import StationSet

logger = logging.getLogger("opennem.importer.gi")

participant_keys = ["name", "abn"]

FACILITY_KEYS = [
    "Region",
    "Status",
    "station_name",
    "owner",
    "TechType",
    "FuelType",
    "duid",
    "unit_num",
    "unit_capacity_lower",
    "unit_capacity",
    "capacity_upper",
    "capacity_registered",
    "capacity_storage",
    "status",
    "dispatch_type",
    "UseDate",
    "ClosureDateExpected",
    "ClosureDate",
    "SurveyID",
    "SurveyVersion",
    "SurveyEffective",
]


def parse_comissioned_date(
    date_str: Union[str, datetime]
) -> Optional[datetime]:
    dt = None

    if type(date_str) is datetime:
        dt = date_str

    try:
        if type(date_str) is str:
            dt = datetime.strptime(date_str, "%d/%m/%y")
    except ValueError:
        logger.error("Error parsing date: {}".format(date_str))

    return dt


def lookup_station_code(
    duids: List[str], station_name: str, station_code_map: dict
) -> Optional[str]:

    station_code = None
    station_name = station_name.strip()

    for facility_duid in duids:
        if facility_duid in station_code_map:
            station_code = station_code_map[facility_duid]
            continue

    if not station_code:
        station_code = station_name_code_map(station_name)

    if not station_code:
        station_code = station_code_from_duids(duids)

        logger.info(
            "Had to generate station code from duids: {} from {}".format(
                station_code, ",".join(duids)
            )
        )

    return station_code


FACILITY_INVALID_STATUS = [
    "Publically Announced",
    "Upgrade",
    "Emerging",
    "Expansion",
    "Maturing",
]


def gi_filter(record):
    if record["station_name"] is None:
        return False

    if record["FuelType"] in ["Natural Gas Pipeline"]:
        return False

    # skip these statuses too
    if record["status"] in FACILITY_INVALID_STATUS:
        return False

    return True


capacity_keys = [
    "unit_capacity_lower",
    "unit_capacity",
    "capacity_upper",
    "capacity_registered",
    "capacity_storage",
]


def get_capacities(record) -> dict:
    _d = {}

    for capacity_key in capacity_keys:
        _d[capacity_key] = clean_capacity(record[capacity_key])

    return _d


def gi_grouper(records, station_code_map):

    # filter out records we don't want
    records = list(filter(gi_filter, records))

    records = [
        {"name": station_name_cleaner(i["station_name"]), **i} for i in records
    ]

    grouped_records = {}

    for k, v in groupby(
        records, key=lambda v: (v["name"].strip(), v["owner"].strip())
    ):
        v = list(v)

        key = k[0]

        if key not in grouped_records:
            grouped_records[key] = []

        grouped_records[key] += v

    records_parsed = []

    for station_name, facilities in grouped_records.items():
        facility_index = 0

        for i in facilities:
            name = station_name_cleaner(i["station_name"])
            duid = normalize_duid(i["duid"])

            unit = parse_unit_duid(i["unit_num"], duid)
            units_num = i["unit_num"] or 1
            unit_id = facility_index + (units_num - 1)
            unit = parse_unit_duid(unit_id, duid)

            fueltech = lookup_fueltech(i["FuelType"], techtype=i["TechType"])

            if not duid:
                duid = get_unit_code(unit, duid, i["station_name"])

            facility_duids = [duid]
            station_code = lookup_station_code(
                facility_duids, station_name, station_code_map
            )

            records_parsed.append(
                {
                    # not a real station id
                    # "id": _id,
                    "name": name,
                    "code": duid,
                    "network_code": duid,
                    "station_code": station_code,
                    "network_region": normalize_aemo_region(i["Region"]),
                    "network_name": i["station_name"].strip(),
                    "dispatch_type": "GENERATOR",
                    "fueltech": parse_facility_fueltech(fueltech)
                    if fueltech
                    else None,
                    "status": parse_facility_status(
                        map_aemo_facility_status(i["status"])
                    ),
                    "registered": parse_comissioned_date(i["SurveyEffective"]),
                    **get_capacities(i),
                }
            )

            facility_index += 1

    coded_records = {}
    _id = 3000

    for station_code, facilities in groupby(
        records_parsed, key=lambda x: x["station_code"]
    ):

        # station_name = facilities[0]["name"]
        facilities = list(facilities)

        if not station_code:
            raise Exception(
                "Unmapped station {}: {}".format(station_code, facilities)
            )

        if station_code not in coded_records:
            coded_records[station_code] = {
                "id": _id,
                "code": station_code,
                # "name": station_name_cleaner(station_name),
                # "network_name": station_name,
                "facilities": [],
            }
            _id += 1

        coded_records[station_code]["facilities"] += facilities

    grouped_records = {}

    for station_code, station_record in coded_records.items():
        grouped_records[station_code] = {
            "id": station_record.get("id"),
            "code": station_code,
            "name": station_name_cleaner(station_name),
            "network_name": station_name,
            "facilities": [],
        }

        for facility in station_record["facilities"]:
            grouped_records[station_code]["name"] = facility["name"]
            grouped_records[station_code]["network_name"] = facility[
                "network_name"
            ]
            facility.pop("name")
            facility.pop("network_name")
            grouped_records[station_code]["facilities"].append(facility)

    return grouped_records


def load_gi():
    aemo_path = PROJECT_DATA_PATH / "aemo" / "nem_gi_202007.xlsx"

    if not aemo_path.is_file():
        raise Exception("Not found: {}".format(aemo_path))

    wb = load_workbook(aemo_path, data_only=True)

    generator_ws = wb["ExistingGeneration&NewDevs"]

    records = []

    for row in generator_ws.iter_rows(min_row=3, values_only=True):
        row_collapsed = row[0:10] + (row[11],) + row[12:19] + row[21:]

        return_dict = dict(zip(FACILITY_KEYS, list(row_collapsed)))

        records.append(return_dict)

    return records


def gi_import():
    mms_duid_station_map = load_data("mms_duid_station_map.json", True)

    nem_gi = load_gi()
    nem_gi = gi_grouper(nem_gi, mms_duid_station_map)

    gi = StationSet()

    for r in nem_gi.values():
        gi.add_dict(r)

    return gi


def gi_export():
    nem_gi = gi_import()

    with open("data/nem_gi.json", "w") as fh:
        fh.write(nem_gi.json(indent=4))

    logger.info("Wrote {} records".format(nem_gi.length))


if __name__ == "__main__":
    gi_export()



import re

from opennem.pipelines.bulk_insert import BulkInsertPipeline
from opennem.pipelines.csv import RecordsToCSVPipeline
from opennem.pipelines.nem.opennem import NemwebUnitScadaOpenNEMStorePipeline
from opennem.spiders.nemweb import NemwebSpider


class NemwebLatestRooftopActual(NemwebSpider):
    name = "au.nem.latest.rooftop"
    start_url = "http://www.nemweb.com.au/Reports/CURRENT/ROOFTOP_PV/ACTUAL/"
    limit = 2

    # Ignore sat
    filename_filter = re.compile(r".*_MEASUREMENT_.*")

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            BulkInsertPipeline,
            RecordsToCSVPipeline,
        ]
    )


class NemwebCurrentRooftopActual(NemwebSpider):
    name = "au.nem.current.rooftop"
    start_url = "http://www.nemweb.com.au/Reports/CURRENT/ROOFTOP_PV/ACTUAL/"

    filename_filter = re.compile(r".*_MEASUREMENT_.*")

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            BulkInsertPipeline,
            RecordsToCSVPipeline,
        ]
    )


class NemwebArchiveRooftopActual(NemwebSpider):
    name = "au.nem.archive.rooftop"
    start_url = "http://www.nemweb.com.au/Reports/ARCHIVE/ROOFTOP_PV/ACTUAL/"

    filename_filter = re.compile(r".*_MEASUREMENT_.*")

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            BulkInsertPipeline,
            RecordsToCSVPipeline,
        ]
    )



# pylint: disable=no-member
"""
facility + station unique constraint on code + network

Revision ID: eeb81f56767e
Revises: 4bf86ff5c8ff
Create Date: 2020-11-24 11:03:42.278896

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "eeb81f56767e"
down_revision = "4bf86ff5c8ff"
branch_labels = None
depends_on = None


def upgrade():
    # clean up stations + facilities from dupe bug
    op.execute(
        "delete from facility where id not in (select min(id) from facility group by network_id, code)"
    )
    op.execute("delete from facility where approved is false")
    op.execute(
        "delete from station where id not in (select min(id) from station group by code)"
    )

    # facility table
    op.alter_column("facility", "code", nullable=False)
    op.create_unique_constraint(
        "excl_facility_network_id_code", "facility", ["network_id", "code"]
    )

    # station table constraints
    op.alter_column("station", "code", nullable=False)
    op.create_unique_constraint(
        "excl_station_network_duid", "station", ["code"]
    )


def downgrade():
    op.alter_column("facility", "code", nullable=True)
    op.drop_constraint("excl_facility_network_id_code")

    op.alter_column("station", "code", nullable=True)
    op.drop_constraint("excl_station_network_duid")



import re
from logging import Filter
from typing import Optional

import yaml

from opennem.core.loader import load_data

DEFAULT_LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "loggers": {
        "opennem": {
            "level": "DEBUG",
        },
        "scrapy": {
            "level": "ERROR",
        },
        "twisted": {
            "level": "ERROR",
        },
    },
}


class SettingsNotFound(Exception):
    pass


def load_logging_config(
    filename: str = "logging.yml", fail_silent: bool = True
) -> Optional[dict]:
    """"""

    settings_file_content = load_data(filename, from_settings=True)

    if not settings_file_content:
        if fail_silent:
            return None

        raise SettingsNotFound(
            "Not a valid logging settings file: {}".format(filename)
        )

    config_data = yaml.safe_load(settings_file_content)

    return config_data


LOGGING_CONFIG = load_logging_config()

class ItemMessageFilter(Filter):
    def filter(self, record):
        # The message that logs the item actually has raw % operators in it,
        # which Scrapy presumably formats later on
        match = re.search(
            r"(Scraped from %\(src\)s)\n%\(item\)s", record.msg
        )
        if match:
            # Make the message everything but the item itself
            record.msg = match.group(1)
        # Don't actually want to filter out this record, so always return 1
        return 1



# pylint: disable=no-member
"""
Facility Scada and Balancing Summary facility and network indicies

Revision ID: 72a817116434
Revises: 467a3223b204
Create Date: 2020-12-07 23:57:33.383834

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "72a817116434"
down_revision = "467a3223b204"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        "create index if not exists idx_facility_scada_facility_code_trading_interval on facility_scada (facility_code, trading_interval desc)"
    )

    op.execute(
        "create index if not exists idx_facility_scada_network_id_trading_interval on facility_scada (network_id, trading_interval desc)"
    )

    op.execute(
        "create index if not exists idx_balancing_summary_network_id_trading_interval on balancing_summary (network_id, trading_interval desc)"
    )

    op.execute(
        "create index if not exists idx_balancing_summary_network_region_trading_interval on balancing_summary (network_region, trading_interval desc);"
    )


def downgrade() -> None:
    op.execute(
        "drop index if exists idx_facility_scada_facility_code_trading_interval"
    )
    op.execute(
        "drop index if exists idx_facility_scada_network_id_trading_interval"
    )
    op.execute(
        "drop index if exists idx_balancing_summary_network_id_trading_interval"
    )
    op.execute(
        "drop index if exists idx_balancing_summary_network_region_trading_interval"
    )



"""
OpenNEM Parser for Australian CPI Data

Source: https://www.rba.gov.au/statistics/tables/xls/g01hist.xls?v=2020-12-24-16-59-11

"""

import logging
from datetime import datetime
from typing import List

import xlrd
from pydantic.error_wrappers import ValidationError

from opennem.schema.stats import AUCpiData, StatDatabase, StatsSet, StatTypes
from opennem.utils.http import http

logger = logging.getLogger("opennem.stats.au.api")

AU_CPI_URL = "https://www.rba.gov.au/statistics/tables/xls/g01hist.xls"


def fetch_au_cpi() -> List[AUCpiData]:
    """Gets australian CPI figures and parses into JSON"""
    r = http.get(AU_CPI_URL)

    if not r.ok:
        raise Exception(
            "Problem grabbing CPI source: {}".format(r.status_code)
        )

    wb = xlrd.open_workbook(file_contents=r.content)

    if "Data" not in wb.sheet_names():
        raise Exception("No Data sheet in CPI workbook")

    wb_data = wb.sheet_by_index(0)
    records = []

    for i in range(11, wb_data.nrows):
        row = wb_data.row_values(i, 0, 2)

        cpi_record = None

        # skip empty values
        if not row[1]:
            continue

        try:
            cpi_record = AUCpiData(quarter_date=row[0], cpi_value=row[1])
        except ValidationError as e:
            logger.info("Invalid CPI data: {}".format(e))
            continue

        if cpi_record:
            records.append(cpi_record)

    return records


def au_cpi_to_statset(records: List[AUCpiData]) -> StatsSet:

    stat_set = [
        StatDatabase(
            stat_date=i.quarter_date,
            stat_type=StatTypes.CPI,
            value=i.cpi_value,
        )
        for i in records
    ]

    s = StatsSet(
        name="au.cpi",
        source_url=AU_CPI_URL,
        fetched_date=datetime.now(),
        stats=stat_set,
    )

    return s


def stat_au_cpi() -> StatsSet:
    records = fetch_au_cpi()
    return au_cpi_to_statset(records)


if __name__ == "__main__":
    r = stat_au_cpi()



from opennem.pipelines.bulk_insert import BulkInsertPipeline
from opennem.pipelines.csv import RecordsToCSVPipeline
from opennem.pipelines.nem.opennem import NemwebUnitScadaOpenNEMStorePipeline
from opennem.spiders.nemweb import NemwebSpider


class NemwebDispatchActualGenLatest(NemwebSpider):
    name = "au.nem.latest.dispatch_actual_gen"
    start_url = "http://www.nemweb.com.au/Reports/CURRENT/Next_Day_Actual_Gen/"

    limit = 1

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            BulkInsertPipeline,
            RecordsToCSVPipeline,
        ]
    )


class NemwebDispatchActualGenCurrent(NemwebSpider):
    name = "au.nem.current.dispatch_actual_gen"
    start_url = "http://www.nemweb.com.au/Reports/CURRENT/Next_Day_Actual_Gen/"

    limit = 0
    skip = 1

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            BulkInsertPipeline,
            RecordsToCSVPipeline,
        ]
    )


class NemwebDispatchActualGenArchive(NemwebSpider):
    name = "au.nem.archive.dispatch_actual_gen"
    start_url = "http://www.nemweb.com.au/Reports/ARCHIVE/Next_Day_Actual_Gen/"
    limit = 0

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            BulkInsertPipeline,
            RecordsToCSVPipeline,
        ]
    )

    # Archives tend to contain large zips of embedded zips so throttle
    # to limit memory use
    custom_settings = {
        "CONCURRENT_REQUESTS": 1,
        "CONCURRENT_ITEMS": 1,
    }



from opennem.db import get_database_engine
from opennem.utils.dates import subtract_days

TIMESCALE_VIEWS = ["mv_facility_energy_hour", "mv_nem_facility_power_5min"]

MATERIAL_VIEWS = ["mv_facility_all"]


def refresh_timescale_views() -> None:
    """refresh timescale views"""
    __query = """
    CALL refresh_continuous_aggregate(
        continuous_aggregate => '{view}',
        window_start         => '{date_from}',
        window_end           => NULL
    );
    """

    engine = get_database_engine()

    dt = subtract_days(days=7)
    date_from = dt.strftime("%Y-%m-%d")

    with engine.connect() as c:
        for v in TIMESCALE_VIEWS:
            query = __query.format(view=v, date_from=date_from)
            print(query)
            c.execution_options(isolation_level="AUTOCOMMIT").execute(query)


def refresh_material_views() -> None:
    """Refresh material views"""
    __query = "REFRESH MATERIALIZED VIEW CONCURRENTLY {view} with data"

    engine = get_database_engine()

    with engine.connect() as c:
        for v in MATERIAL_VIEWS:
            query = __query.format(view=v)
            c.execution_options(isolation_level="AUTOCOMMIT").execute(query)


def refresh_views() -> None:
    # refresh_timescale_views()
    refresh_material_views()


if __name__ == "__main__":
    refresh_timescale_views()



# pylint: disable=no-member
"""Add is_forecast to scada and balancing

Revision ID: 458bf49bd781
Revises: b16d7153a221
Create Date: 2020-10-24 00:55:23.513798

"""
import sqlalchemy as sa
from alembic import op

revision = "458bf49bd781"
down_revision = "b16d7153a221"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "balancing_summary",
        sa.Column("is_forecast", sa.Boolean(), nullable=True),
    )
    op.add_column(
        "facility_scada", sa.Column("is_forecast", sa.Boolean(), nullable=True)
    )


def downgrade():
    op.drop_column("facility_scada", "is_forecast")
    op.drop_column("balancing_summary", "is_forecast")



"""
    Parse AEMO CSV format which can have multiple
    tables and definitions per CSV files.

"""

import csv
import logging
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, validator
from pydantic.error_wrappers import ValidationError
from pydantic.fields import PrivateAttr

from opennem.schema.aemo.mms import get_mms_schema_for_table

logger = logging.getLogger(__name__)


# pylint: disable=no-self-argument
class AEMOTableSchema(BaseModel):
    name: str
    namespace: str
    fieldnames: List[str]
    records: List[Union[Dict, BaseModel]] = []

    # optionally it has a schema
    _record_schema: Optional[BaseModel] = PrivateAttr()

    # the url this table was taken from if any
    url_source: Optional[str]

    # the original content ?!?!
    # @NOTE does this make sense ..
    content_source: Optional[str]

    @property
    def full_name(self) -> str:
        return "{}_{}".format(self.namespace, self.name)

    @validator("name")
    def validate_name(cls, table_name: str) -> str:
        _table_name = table_name.strip().upper()

        return _table_name

    @validator("namespace")
    def validate_namespace(cls, namespace_name: str) -> str:
        _namespace_name = namespace_name.strip().upper()

        return _namespace_name

    @validator("fieldnames")
    def validate_fieldnames(cls, fieldnames: List[str]) -> List[str]:
        if not isinstance(fieldnames, list):
            return []

        _fieldnames = [i.strip() for i in fieldnames]

        return _fieldnames

    def set_schema(self, schema: BaseModel) -> bool:
        self._record_schema = schema

        return True

    def get_records(self) -> Any:
        return self.records

    def add_record(self, record: Union[Dict, BaseModel]) -> bool:
        if self._record_schema:
            _record = None

            try:
                _record = self._record_schema(**record)  # type: ignore
            except ValidationError as e:
                val_errors = e.errors()

                for ve in val_errors:
                    ve_fieldname = ve["loc"][0]
                    ve_val = ""

                    if (
                        record
                        and isinstance(record, dict)
                        and ve_fieldname in record
                    ):
                        ve_val = record[ve_fieldname]

                    logger.error(
                        "{} has error: {} '{}'".format(
                            ve_fieldname, ve["msg"], ve_val
                        )
                    )
                return False

            self.records.append(_record)
        else:
            self.records.append(record)

        return True

    class Config:
        underscore_attrs_are_private = True


class AEMOTableSet(BaseModel):
    tables: List[AEMOTableSchema] = []

    def has_table(self, table_name: str) -> bool:
        if not self.tables:
            return False

        if len(self.tables) < 1:
            return False

        table_name = table_name.upper()

        table_lookup = list(
            filter(lambda t: t.full_name == table_name, self.tables)
        )

        return len(table_lookup) > 0

    def add_table(self, table: AEMOTableSchema) -> bool:
        self.tables.append(table)

        return True

    def get_table(self, table_name: str) -> AEMOTableSchema:
        table_name = table_name.upper()

        if not self.has_table(table_name):
            raise Exception("Table not found: {}".format(table_name))

        table_lookup = list(
            filter(lambda t: t.full_name == table_name, self.tables)
        )

        return table_lookup.pop()


class AEMOParserException(Exception):
    pass


AEMO_ROW_HEADER_TYPES = ["C", "I", "D"]


def parse_aemo_csv(content: str) -> AEMOTableSet:
    """
    Parse AEMO CSV's into schemas and return a table set

    Exception raised on error and logs malformed CSVs
    """
    content_split = content.splitlines()

    datacsv = csv.reader(content_split)

    # init all the parser vars
    table_set = AEMOTableSet()
    table_current = None

    for row in datacsv:
        if not row or type(row) is not list or len(row) < 1:
            continue

        record_type = row[0].strip().upper()

        if record_type not in AEMO_ROW_HEADER_TYPES:
            logger.info("Skipping row, invalid type: {}".format(record_type))
            continue

        # new table set
        if record_type == "C":
            # @TODO csv meta stored in table
            if table_current:
                table_set.add_table(table_current)

        # new table
        elif record_type == "I":
            if table_current:
                table_set.add_table(table_current)

            table_namespace = row[1]
            table_name = row[2]
            table_fields = row[4:]

            table_current = AEMOTableSchema(
                name=table_name,
                namespace=table_namespace,
                fields=table_fields,
                fieldnames=table_fields,
            )

            # do we have a custom shema for the table?
            table_schema = get_mms_schema_for_table(table_current.full_name)

            if table_schema:
                table_current.set_schema(table_schema)

        # new record
        elif record_type == "D":
            if not table_current:
                logger.error(
                    "Malformed AEMO csv - field records before table definition"
                )
                continue

            values = row[4:]

            if len(values) != len(table_current.fieldnames):
                logger.error(
                    "Malformed AEMO csv - length mismatch between records and fields"
                )
                continue

            record = dict(zip(table_current.fieldnames, values))

            table_current.add_record(record)

    return table_set



"""
Profiler - profile code in context

Usage:

```py
with profiled():
    code_to_run()
```
"""

import contextlib
import cProfile
import io
import pstats


@contextlib.contextmanager
def profiled():
    pr = cProfile.Profile()
    pr.enable()
    yield
    pr.disable()
    s = io.StringIO()
    ps = pstats.Stats(pr, stream=s).sort_stats("cumulative")
    ps.print_stats()
    # uncomment this to see who's calling what
    # ps.print_callers()
    print(s.getvalue())



import logging
from collections import OrderedDict
from datetime import datetime, timezone
from textwrap import dedent
from typing import Any, Dict, List, Optional, Union

from sqlalchemy.orm import Session

from opennem.core.normalizers import normalize_duid
from opennem.core.units import get_unit
from opennem.db import get_database_engine
from opennem.db.models.opennem import FacilityScada, Station
from opennem.schema.network import NetworkSchema
from opennem.schema.time import TimeInterval, TimePeriod
from opennem.schema.units import UnitDefinition
from opennem.utils.cache import cache_scada_result
from opennem.utils.numbers import cast_trailing_nulls
from opennem.utils.time import human_to_timedelta
from opennem.utils.timezone import is_aware, make_aware
from opennem.utils.version import get_version

from .schema import (
    DataQueryResult,
    OpennemData,
    OpennemDataHistory,
    OpennemDataSet,
    ScadaDateRange,
)

logger = logging.getLogger(__name__)


def duid_in_case(facility_codes: List[str]) -> str:
    """Takes a list of duids and returns a formatted sql array for IN"""
    map_list = ["'{}'".format(i) for i in map(normalize_duid, facility_codes)]
    return ",".join(map_list)


def stats_factory(
    stats: List[DataQueryResult],
    units: UnitDefinition,
    interval: Optional[TimeInterval] = None,
    period: Optional[TimePeriod] = None,
    network: Optional[NetworkSchema] = None,
    timezone: Optional[Union[timezone, str]] = None,
    code: Optional[str] = None,
    region: Optional[str] = None,
    fueltech_group: Optional[bool] = False,
    group_field: Optional[str] = None,
    data_id: Optional[str] = None,
) -> Optional[OpennemDataSet]:
    """
    Takes a list of data query results and returns OpennemDataSets

    @TODO optional groupby field
    @TODO override timezone
    @TODO multiple groupings / slight refactor

    """

    if network:
        timezone = network.get_timezone()

    dates = [s.interval for s in stats]

    if not dates:
        return None

    start = min(dates)
    end = max(dates)

    # should probably make sure these are the same TZ
    if timezone and not is_aware(start):
        start = make_aware(min(dates), timezone)

    if timezone and not is_aware(end):
        end = make_aware(max(dates), timezone)

    # free
    dates = []

    group_codes = list(set([i.group_by for i in stats if i.group_by]))

    stats_grouped = []

    for group_code in group_codes:

        data_grouped: Dict[Any, Any] = dict()

        for stat in stats:
            if stat.interval not in data_grouped:
                data_grouped[stat.interval] = None

            if stat.group_by == group_code:
                data_grouped[stat.interval] = stat.result

        data_sorted = OrderedDict(sorted(data_grouped.items()))

        data_value = list(data_sorted.values())

        # Skip null series
        if len([i for i in data_value if i]) == 0:
            continue

        # Skip zero series
        if sum([i for i in data_value if i]) == 0:
            continue

        # Cast trailing nulls
        if not units.name.startswith("temperature") or units.cast_nulls:
            data_value = cast_trailing_nulls(data_value)

        history = OpennemDataHistory(
            start=start,
            last=end,
            data=data_value,
        )

        if interval:
            history.interval = interval.interval_human

        data = OpennemData(
            data_type=units.unit_type,
            units=units.unit,
            code=group_code,
            # interval=interval,
            # period=period,
            history=history,
        )

        if network:
            data.network = network.code

        # *sigh* - not the most flexible model
        # @TODO fix this schema and make it more flexible
        if fueltech_group:
            data.fuel_tech = group_code
            data_comps = [
                # @NOTE disable for now since FE doesn't
                # support it
                # network.country if network else None,
                network.code.lower() if network else None,
                region.lower() if region else None,
                "fuel_tech",
                group_code.lower(),
                units.unit_type,
            ]
            data.id = ".".join(i for i in data_comps if i)
            # @TODO make this an alias
            data.type = units.unit_type

        if group_field:
            group_fields = []

            # setattr(data, group_field, group_code)

            if network:
                group_fields.append(network.code.lower())

            group_fields = group_fields + [
                group_field,
                group_code,
                units.unit_type,
            ]

            data.id = ".".join([f for f in group_fields if f])
            data.type = units.unit_type

        if data_id:
            data.id = data_id

        if not data.id:
            _id_list = []

            if region:
                _id_list.append(region.lower())

            elif network:
                _id_list.append(network.code)

            if units and units.name_alias:
                _id_list.append(units.name_alias)

            data.id = ".".join([f.lower() for f in _id_list if f])
            data.type = units.unit_type

        if region:
            data.region = region

        stats_grouped.append(data)

    dt_now = datetime.now()

    if network:
        dt_now = dt_now.astimezone(network.get_timezone())

    # @NOTE this should probably be
    # country.network.region
    if not code:
        if network:
            code = network.code.lower()

        if region:
            code = region.lower()

    stat_set = OpennemDataSet(
        type=units.unit_type,
        data=stats_grouped,
        created_at=dt_now,
        code=code,
        version=get_version(),
    )

    if network:
        stat_set.network = network.code

    if region:
        stat_set.region = region

    return stat_set


def networks_to_in(networks: List[NetworkSchema]) -> str:
    codes = ["'{}'".format(n.code) for n in networks]

    return ", ".join(codes)


SCADA_RANGE_EXCLUDE_DUIDS = [
    "V-SA",
    "V-SA-2",
    "N-Q-MNSP1",
    "N-Q-MNSP1-2",
    "NSW1-QLD1",
    "NSW1-QLD1-2",
    "V-S-MNSP1",
    "V-S-MNSP1-2",
    "T-V-MNSP1",
    "T-V-MNSP1-2",
    "VIC1-NSW1",
    "VIC1-NSW1-2",
]


@cache_scada_result
def get_scada_range(
    network: Optional[NetworkSchema] = None,
    networks: Optional[List[NetworkSchema]] = None,
    network_region: Optional[str] = None,
    facilities: Optional[List[str]] = None,
) -> Optional[ScadaDateRange]:
    """Get the start and end dates for a network query. This is more efficient
    than providing or querying the range at query time
    """
    engine = get_database_engine()

    __query = """
        select
            min(fs.trading_interval)::timestamp AT TIME ZONE '{timezone}',
            max(fs.trading_interval)::timestamp AT TIME ZONE '{timezone}'
        from facility_scada fs
        where
            {facility_query}
            {network_query}
            {network_region_query}
            facility_code not like 'ROOFTOP_%%'
            and facility_code not in ({exclude_duids})
    """

    network_query = ""
    timezone = "UTC"

    if network:
        network_query = f"fs.network_id = '{network.code}' and"
        # timezone = network.timezone_database

    if networks:
        net_case = networks_to_in(networks)
        network_query = "fs.network_id IN ({}) and ".format(net_case)

    network_region_query = ""

    if network_region:
        # @TODO support network regions in scada_range
        # technically we don't need this atm
        pass
        # network_region_query = (
        # f"f.network_region = '{network_region.code}' and"
        # )

    facility_query = ""

    if facilities:
        fac_case = duid_in_case(facilities)
        facility_query = "fs.facility_code IN ({}) and ".format(fac_case)

    exclude_duids = duid_in_case(SCADA_RANGE_EXCLUDE_DUIDS)

    scada_range_query = dedent(
        __query.format(
            facility_query=facility_query,
            network_query=network_query,
            network_region_query=network_region_query,
            timezone=timezone,
            exclude_duids=exclude_duids,
        )
    )

    with engine.connect() as c:
        logger.debug(scada_range_query)
        scada_range_result = list(c.execute(scada_range_query))

        if len(scada_range_result) < 1:
            raise Exception("No scada range result")

        scada_min = scada_range_result[0][0]
        scada_max = scada_range_result[0][1]

    if not scada_min or not scada_max:
        return None

    scada_range = ScadaDateRange(start=scada_min, end=scada_max, network=network)

    return scada_range


def station_attach_stats(station: Station, session: Session) -> Station:
    # @TODO update for new queries
    since = datetime.now() - human_to_timedelta("7d")

    facility_codes = list(set([f.code for f in station.facilities]))

    stats = (
        session.query(FacilityScada)
        .filter(FacilityScada.facility_code.in_(facility_codes))
        .filter(FacilityScada.trading_interval >= since)
        .order_by(FacilityScada.facility_code)
        .order_by(FacilityScada.trading_interval)
        .all()
    )

    for facility in station.facilities:
        facility_power = list(filter(lambda s: s.facility_code == facility.code, stats))

        # facility.scada_power = stats_factory(facility_power)

    return station



from io import BytesIO
from itertools import groupby

import scrapy
from openpyxl import load_workbook


class XLSSpider(scrapy.Spider):
    def start_requests(self):
        if not hasattr(self, "url"):
            raise Exception("Require URL param")

        yield scrapy.Request(self.url)

    def parse(self, response):
        wb = load_workbook(BytesIO(response.body), data_only=True)

        ws = wb.get_sheet_by_name("ExistingGeneration&NewDevs")

        for row in ws.iter_rows(min_row=3, values_only=True):

            # pick out the columns we want
            # lots of hidden columns in the sheet
            row_collapsed = row[0:8] + (row[11],) + row[12:19] + row[22:]

            return_dict = dict(zip(self.keys, list(row_collapsed)))

            if return_dict is None:
                raise Exception("Failed on row: {}".format(row))

            yield return_dict



# pylint: disable=no-member
"""Create bom_observation hypertable

Revision ID: bb6fb4645ca3
Revises: 4ee58f229572
Create Date: 2020-10-21 14:20:02.009394

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "bb6fb4645ca3"
down_revision = "4ee58f229572"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        "select create_hypertable('bom_observation', 'observation_time', if_not_exists => TRUE, migrate_data => TRUE, chunk_time_interval => INTERVAL '1 week')"
    )


def downgrade():
    pass



import json
from operator import attrgetter
from typing import Any, List

from opennem.core.loader import load_data
from opennem.core.normalizers import clean_capacity, normalize_whitespace
from opennem.importer.compat import (
    map_compat_facility_state,
    map_compat_fueltech,
    map_compat_network_region,
)

from .schema import FacilitySchema, StationSchema

__all__ = ["load_registry", "load_current"]


def _sort_stations(records: List[Any]) -> List[Any]:
    return sorted(records, key=attrgetter("state", "name"))


def _sort_facilities(records: List[Any]) -> List[Any]:
    return sorted(records, key=attrgetter("duid", "status"))


def _get_state_from_current(station, facilities: List[Any]) -> str:
    state = station.get("state")

    if not state:
        facility_states = list(
            set([f.network_region for f in facilities if f.network_region])
        )

        state = facility_states.pop()

    state = state.upper()

    if state == "WEM":
        state = "WA"

    if state.endswith("1"):
        state = state[:-1]

    return state


def load_registry() -> List[StationSchema]:
    """
        Loads the facility registry into a list of Station schema's
    """
    stations = load_data("facility_registry.json")

    records = []

    for station_id, station_record in stations.items():
        facilities = []

        for duid, facility_record in station_record["duid_data"].items():
            status = map_compat_facility_state(
                station_record.get("status", {}).get("state", "")
            )
            fuel_tech = map_compat_fueltech(
                facility_record.get("fuel_tech", "")
            )
            registered_capacity = clean_capacity(
                facility_record.get("registered_capacity")
            )

            facility = FacilitySchema(
                name=normalize_whitespace(station_record.get("display_name")),
                network_region=map_compat_network_region(
                    station_record["region_id"]
                ),
                status=status,
                duid=duid,
                fueltech=fuel_tech,
                capacity=registered_capacity,
            )

            facilities.append(facility)

        record = StationSchema(
            name=normalize_whitespace(station_record.get("display_name")),
            code=station_id,
            state=station_record.get("location", {}).get("state", None),
            facilities=_sort_facilities(facilities),
        )

        records.append(record)

    records = _sort_stations(records)

    return records


def load_current() -> List[StationSchema]:
    """
        Load the current project station data into a list of station schemas
    """
    station_data = load_data("stations.geojson", True)

    records = []

    for s in station_data.get("features", []):
        station = s.get("properties")

        facilities = []

        for facility in station.get("duid_data", {}):

            facility = FacilitySchema(
                name=station.get("name"),
                network_region=facility.get("network_region"),
                status=facility.get("status"),
                duid=facility.get("duid"),
                fueltech=facility.get("fuel_tech"),
                capacity=facility.get("capacity_registered"),
            )

            facilities.append(facility)

        record = StationSchema(
            name=station.get("name"),
            code=station.get("station_code"),
            state=_get_state_from_current(station, facilities),
            facilities=_sort_facilities(facilities),
        )

        records.append(record)

    records = _sort_stations(records)

    return records



import logging
import re
from datetime import datetime
from pprint import pprint
from typing import List, Optional, Union

from dictalchemy.utils import fromdict
# from opennem.core.loader import load_data
from pydantic import BaseModel
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import MultipleResultsFound

from opennem.core.loader import load_data
from opennem.db import SessionLocal
from opennem.db.load_fixtures import load_fixtures
from opennem.db.models.opennem import Facility, Location, Revision, Station
from opennem.importer.aemo_gi import gi_import
from opennem.importer.aemo_rel import rel_import
from opennem.importer.emissions import import_emissions_csv
from opennem.importer.mms import mms_import
from opennem.importer.registry import registry_import
from opennem.schema.stations import StationSet

from .comparator import compare_record_differs

logger = logging.getLogger(__name__)
s = SessionLocal()


def load_revisions(stations):
    pass


def create_revision(model):
    # s = session()

    # make_transient(model)

    clone = model.asdict(exclude_pk=True)
    # clone.pop("id")
    # model.fromdict(clone)

    # pylint: disable=no-member
    clone_model = Station().fromdict(clone)

    return clone_model


def get_schema_name(record: object) -> str:
    class_name = str(record.__class__.__name__)

    class_name = re.sub("Schema", "", class_name)

    class_name = class_name.lower()

    return class_name


def revision_factory_join_field(record_instance: BaseModel) -> str:
    class_name = str(record_instance.__class__.__name__)

    FIELD_MAP = {
        "StationSchema": "station_id",
        "FacilitySchema": "facility_id",
        "LocationSchema": "location_id",
    }

    if class_name in FIELD_MAP.keys():
        return FIELD_MAP[class_name]

    raise Exception("Could not map record %s", record_instance)


def revision_factory(
    record: BaseModel,
    field_names: Union[str, List[str]],
    created_by: str,
) -> Optional[Revision]:

    if isinstance(field_names, str):
        field_names = [field_names]

    revision_data = {}

    for field_name in field_names:
        # field_type = get_schema_name(record)
        field_value = getattr(record, field_name)

        if isinstance(field_value, BaseModel):
            _value_dict = field_value.dict()

            # we only store the primary keys in associations rather
            # than the full record.
            if "id" in _value_dict:
                field_value = _value_dict["id"]

            elif "code" in _value_dict:
                field_value = _value_dict["code"]

            else:
                logger.error("Could not serialize data value %s", field_value)
                field_value = False

        if field_value is not False:
            revision_data[field_name] = field_value

    if not record.id:
        logger.error("Require a code to create a revision: %s", record)
        return None

    if len(revision_data.keys()) < 1:
        logger.error("Require data to lookup for code %s", record.code)
        return None

    column_name = revision_factory_join_field(record)

    __query = (
        s.query(Revision).filter(getattr(Revision, column_name) == record.id)
        # .filter(Revision.code == record.code)
    )

    for data_name, data_value in revision_data.items():
        if isinstance(data_value, str):
            __query = __query.filter(
                Revision.changes[data_name].as_string() == data_value
            )
        if isinstance(data_value, bool):
            __query = __query.filter(
                Revision.changes[data_name].as_boolean() == data_value
            )
        if isinstance(data_value, int):
            __query = __query.filter(
                Revision.changes[data_name].as_integer() == data_value
            )
        if isinstance(data_value, float):
            __query = __query.filter(
                Revision.changes[data_name].as_float() == data_value
            )

    revision_lookup = None

    try:
        revision_lookup = __query.one_or_none()
    except MultipleResultsFound:
        logger.info(
            "Revision exists: %s %s %s",
            record.code,
            field_name,
            field_value,
        )
        return None

    if revision_lookup:
        return None

    revision = Revision(
        created_by=created_by,
        changes=revision_data,
    )

    # s.add(revision)
    # s.commit()

    return revision


def load_revision(records, created_by):
    logger.info("Running db test")

    for station_record in records:
        station_model = (
            s.query(Station)
            .filter(Station.code == station_record.code)
            .one_or_none()
        )

        if not station_model:
            logger.info(
                f"New station {station_record.name} {station_record.code}"
            )

            station_dict = station_record.dict(
                include={"code", "network_name", "name", "location"}
            )
            # pprint(station_dict)

            # pylint:disable no-member
            station_model = fromdict(Station(), station_dict)
            station_model.approved = False
            station_model.created_by = created_by

            # location
            if "location" in station_dict and station_dict["location"]:
                station_model.location = fromdict(
                    Location(), station_dict["location"], exclude=["id"]
                )

                if station_record.location.lat and station_record.location.lng:
                    station_model.location.geom = station_record.location.geom

            s.add(station_model)
            s.commit()

            station_record.id = station_model.id

            # revision = revision_factory(
            #     station_record, ["code", "name", "network_name"], created_by,
            # )

            # if revision:
            #     station_model.revisions.append(revision)

            # s.add(station_model)
            # s.commit()

        else:
            for field in ["name"]:
                if getattr(station_model, field) != getattr(
                    station_record, field
                ):
                    revision_factory(station_record, field, created_by)

        for facility in station_record.facilities:
            facility_model = (
                s.query(Facility)
                .filter(Facility.code == facility.code)
                .first()
            )

            if not facility_model:
                logger.info(
                    "New facility %s => %s", station_record.name, facility.code
                )

                facility_dict = facility.dict(
                    include={
                        "code",
                        "network",
                        # "network_id",
                        "dispatch_type",
                        "station",
                        # "station_id",
                        # "status",
                        "network_code",
                        "network_region",
                        "network_name",
                    }
                )
                # pprint(station_dict)

                # pylint:disable no-member
                facility_model = fromdict(Facility(), facility_dict)
                facility_model.network_id = facility.network.code
                facility_model.approved = False
                facility_model.created_by = created_by

                s.add(facility_model)
                s.commit()

                facility.id = facility_model.id

                # @NOTE don't create revisions for new facilities
                # revision = revision_factory(
                #     facility, ["code", "dispatch_type"], created_by
                # )

                # if revision:
                #     facility_model.revisions.append(revision)

                # s.add(facility_model)
                # s.commit()

            else:
                facility.id = facility_model.id

            for field in [
                "fueltech",
                "status",
                "capacity_registered",
            ]:
                revision = None

                if compare_record_differs(facility_model, facility, field):
                    # logger.info(
                    #     "%s and %s differ",
                    #     getattr(facility, field),
                    #     getattr(facility_model, field),
                    # )

                    revision = revision_factory(facility, field, created_by)

                if revision:
                    facility_model.revisions.append(revision)

            s.add(facility_model)

            station_model.facilities.append(facility_model)
            s.add(station_model)
            s.commit()


def db_test():
    s.query(Revision).delete()

    mms = mms_import()
    rel = rel_import()
    gi = gi_import()

    load_revision(mms, "aemo.mms.202006")
    load_revision(rel, "aemo.rel.2020006")
    load_revision(gi, "aemo.gi.202006")


def registry_init() -> None:
    registry = registry_import()
    session = SessionLocal()

    for station in registry:
        station_dict = station.dict(exclude={"id"})
        # pprint(station_dict)

        station_model = (
            session.query(Station).filter_by(code=station.code).one_or_none()
        )

        if not station_model:
            # pylint:disable no-member
            station_model = fromdict(Station(), station_dict)
            station_model.approved = True
            station_model.approved_at = datetime.now()
            station_model.approved_by = "opennem.registry"
            station_model.created_by = "opennem.registry"

        # location
        station_model.location = fromdict(
            Location(),
            station_dict["location"],
            exclude=["id"],
        )

        if station.location.lat and station.location.lng:
            station_model.location.geom = "SRID=4326;POINT({} {})".format(
                station.location.lng, station.location.lat
            )

        session.add(station_model)
        session.commit()

        for fac in station.facilities:
            f = (
                session.query(Facility)
                .filter_by(code=fac.code)
                .filter_by(network_id=fac.network.code)
                .one_or_none()
            )

            if not f:
                print("new facility {} {}".format(fac.code, fac.network.code))
                f = Facility(
                    **fac.dict(
                        exclude={
                            "id",
                            "fueltech",
                            "status",
                            "network",
                            "revision_ids",
                            "scada_power",
                        }
                    )
                )
                f.approved_by = "opennem.registry"
                f.created_by = "opennem.registry"
                f.approved_at = datetime.now()
                f.created_at = datetime.now()

            f.network_id = fac.network.code

            if fac.fueltech:
                f.fueltech_id = fac.fueltech.code

            if fac.network.code:
                f.network_code = fac.network.code

            f.status_id = fac.status.code

            f.approved = True

            if station_model.id:
                f.station_id = station_model.id
            else:
                station_model.facilities.append(f)
                session.add(station_model)

            session.add(f)

        try:
            session.commit()
        except IntegrityError as e:
            logger.error(e)


def opennem_init() -> None:
    session = SessionLocal()
    station_data = load_data("opennem_stations.json", from_project=True)
    stations = StationSet()

    for s in station_data:
        stations.add_dict(s)

    for station in stations:
        logger.debug("Adding station: {}".format(station.code))

        station_model = (
            session.query(Station).filter_by(code=station.code).one_or_none()
        )

        if not station_model:
            station_model = Station(code=station.code)
            station_model.approved = True
            station_model.approved_at = datetime.now()
            station_model.approved_by = "opennem.init"
            station_model.created_by = "opennem.init"

        station_model.description = station.description
        station_model.name = station.name
        station_model.network_name = station.network_name

        if not station_model.location:
            station_model.location = Location()

        if station.location:
            station_model.location.locality = station.location.locality
            station_model.location.state = station.location.state
            station_model.location.postcode = station.location.postcode
            station_model.location.country = station.location.country

        if station.location.lat and station.location.lng:
            station_model.location.geom = "SRID=4326;POINT({} {})".format(
                station.location.lng, station.location.lat
            )

        for fac in station.facilities:
            facility_model = (
                session.query(Facility)
                .filter_by(code=fac.code)
                .filter_by(network_id=fac.network.code)
                .one_or_none()
            )

            if not facility_model:
                facility_model = Facility(
                    code=fac.code, network_id=fac.network.code
                )

            facility_model.network_region = fac.network_region
            facility_model.fueltech_id = fac.fueltech.code
            facility_model.status_id = fac.status.code
            facility_model.dispatch_type = fac.dispatch_type
            facility_model.capacity_registered = fac.capacity_registered
            facility_model.registered = fac.registered
            facility_model.unit_id = fac.unit_id
            facility_model.unit_number = fac.unit_number
            facility_model.unit_alias = fac.unit_alias
            facility_model.unit_capacity = fac.unit_capacity
            facility_model.emissions_factor_co2 = fac.emissions_factor_co2
            facility_model.approved = fac.approved
            facility_model.approved_by = fac.approved_by

            facility_model.created_by = "opennem.init"
            facility_model.approved_by = "opennem.init"

            session.add(facility_model)
            station_model.facilities.append(facility_model)
            logger.debug(" => Added facility {}".format(fac.code))

        print(station_model)
        session.add(station_model)
        session.commit()


def test_revisions() -> None:
    registry = registry_import()
    # mms = mms_import()

    k = registry.get_code("KWINANA")

    station = Station(
        code=k.code,
        name=k.name,
        network_name=k.network_name,
    )
    s.add(station)
    s.commit()

    # pylint: disable=no-member
    station_clone = station.asdict()
    station_clone.pop("id")

    pprint(station_clone)

    station = create_revision(station)

    station.location = Location(
        address1=k.location.address1,
        address2=k.location.address2,
        locality=k.location.locality,
        state=k.location.state,
        postcode=k.location.postcode,
    )

    s.add(station)
    s.commit()

    station = create_revision(station)

    station.name = "Kwinana Edited"

    s.add(station)
    s.commit()


def init() -> None:
    load_fixtures()
    logger.info("Fixture loaded")

    registry_init()
    logger.info("Registry initialized")

    opennem_init()
    logger.info("Opennem initialized")

    import_emissions_csv()
    logger.info("Pollution data initialized")



# pylint: disable=no-member
"""
update facility scada rooftops

Revision ID: 10f179b31ef8
Revises: 15f5aaf2699c
Create Date: 2020-11-17 18:13:43.209415

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "10f179b31ef8"
down_revision = "15f5aaf2699c"
branch_labels = None
depends_on = None


def upgrade():

    op.execute(
        "update facility_scada set facility_code = 'ROOFTOP_APVI_TAS' where facility_code = 'ROOFTOP_TAS'"
    )
    op.execute(
        "update facility_scada set facility_code = 'ROOFTOP_APVI_QLD' where facility_code = 'ROOFTOP_QLD'"
    )
    op.execute(
        "update facility_scada set facility_code = 'ROOFTOP_APVI_NSW' where facility_code = 'ROOFTOP_NSW'"
    )
    op.execute(
        "update facility_scada set facility_code = 'ROOFTOP_APVI_VIC' where facility_code = 'ROOFTOP_VIC'"
    )
    op.execute(
        "update facility_scada set facility_code = 'ROOFTOP_APVI_SA' where facility_code = 'ROOFTOP_SA'"
    )
    op.execute(
        "update facility_scada set facility_code = 'ROOFTOP_APVI_WA' where facility_code = 'ROOFTOP_WA'"
    )
    op.execute(
        "update facility_scada set facility_code = 'ROOFTOP_APVI_NT' where facility_code = 'ROOFTOP_NT'"
    )


def downgrade():

    op.execute(
        "update facility_scada set facility_code = 'ROOFTOP_TAS' where facility_code = 'ROOFTOP_APVI_TAS'"
    )
    op.execute(
        "update facility_scada set facility_code = 'ROOFTOP_QLD' where facility_code = 'ROOFTOP_APVI_QLD'"
    )
    op.execute(
        "update facility_scada set facility_code = 'ROOFTOP_VIC' where facility_code = 'ROOFTOP_APVI_VIC'"
    )
    op.execute(
        "update facility_scada set facility_code = 'ROOFTOP_SA' where facility_code = 'ROOFTOP_APVI_SA'"
    )
    op.execute(
        "update facility_scada set facility_code = 'ROOFTOP_WA' where facility_code = 'ROOFTOP_APVI_WA'"
    )
    op.execute(
        "update facility_scada set facility_code = 'ROOFTOP_NT' where facility_code = 'ROOFTOP_APVI_NT'"
    )



# pylint: disable=no-member
"""BOM additional fields on station and observation

Revision ID: 28a206c84987
Revises: 659f03439631
Create Date: 2020-10-09 05:12:59.186661

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "28a206c84987"
down_revision = "659f03439631"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "bom_observation", sa.Column("cloud", sa.Text(), nullable=True)
    )
    op.add_column(
        "bom_observation", sa.Column("cloud_type", sa.Text(), nullable=True)
    )
    op.add_column(
        "bom_observation", sa.Column("humidity", sa.Numeric(), nullable=True)
    )
    op.add_column(
        "bom_observation", sa.Column("wind_gust", sa.Numeric(), nullable=True)
    )
    op.add_column(
        "bom_station", sa.Column("altitude", sa.Integer(), nullable=True)
    )


def downgrade():
    op.drop_column("bom_station", "altitude")
    op.drop_column("bom_observation", "wind_gust")
    op.drop_column("bom_observation", "humidity")
    op.drop_column("bom_observation", "cloud_type")
    op.drop_column("bom_observation", "cloud")



import logging
from typing import List, Optional

from opennem.api.export.queries import (
    country_stats_query,
    energy_network_fueltech_query,
    interconnector_flow_power_query,
    power_network_fueltech_query,
    price_network_query,
)
from opennem.api.stats.controllers import get_scada_range, stats_factory
from opennem.api.stats.schema import (
    DataQueryResult,
    OpennemDataSet,
    RegionFlowResult,
    ScadaDateRange,
)
from opennem.api.time import human_to_interval, human_to_period
from opennem.api.weather.queries import observation_query, observation_query_all
from opennem.core.flows import net_flows, net_flows_energy
from opennem.core.networks import network_from_network_code
from opennem.core.units import get_unit
from opennem.db import get_database_engine
from opennem.schema.network import NetworkNEM, NetworkSchema
from opennem.schema.stats import StatTypes
from opennem.schema.time import TimePeriod

logger = logging.getLogger(__name__)


def weather_daily(
    network_code: str,
    station_code: str,
    year: Optional[int] = None,
    unit_name: str = "temperature_mean",
    period_human: str = None,
    include_min_max: bool = True,
    date_range: Optional[ScadaDateRange] = None,
) -> Optional[OpennemDataSet]:
    station_codes = []

    if station_code:
        station_codes = [station_code]

    engine = get_database_engine()
    interval = human_to_interval("1d")
    network = network_from_network_code(network_code)
    units = get_unit(unit_name)

    if not date_range:
        date_range = get_scada_range(network=network)

    if not date_range:
        raise Exception("Require a scada range")

    if year:
        query = observation_query(
            station_codes=station_codes,
            interval=interval,
            network=network,
            scada_range=date_range,
            year=year,
        )
    elif period_human:
        interval = human_to_interval("30m")
        period = human_to_period(period_human)

        query = observation_query(
            station_codes=station_codes,
            interval=interval,
            network=network,
            scada_range=date_range,
            period=period,
        )
    else:
        interval = human_to_interval("1M")

        query = observation_query_all(
            station_codes=station_codes,
            scada_range=date_range,
            network=network,
        )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    temp_avg = [
        DataQueryResult(interval=i[0], group_by=i[1], result=i[2] if len(i) > 1 else None)
        for i in row
    ]

    temp_min = [
        DataQueryResult(interval=i[0], group_by=i[1], result=i[3] if len(i) > 1 else None)
        for i in row
    ]

    temp_max = [
        DataQueryResult(interval=i[0], group_by=i[1], result=i[4] if len(i) > 1 else None)
        for i in row
    ]

    if len(temp_avg) < 1:
        logger.info("No results from query: {}".format(query))
        return None

    stats = stats_factory(
        stats=temp_avg,
        units=units,
        network=network,
        interval=interval,
        code="bom",
        group_field="temperature",
    )

    if not stats:
        return None

    if include_min_max:
        stats_min = stats_factory(
            stats=temp_min,
            units=get_unit("temperature_min"),
            network=network,
            interval=interval,
            code="bom",
            group_field="temperature",
        )

        stats_max = stats_factory(
            stats=temp_max,
            units=get_unit("temperature_max"),
            network=network,
            interval=interval,
            code="bom",
            group_field="temperature",
        )

        stats.append_set(stats_min)
        stats.append_set(stats_max)

    return stats


def gov_stats_cpi() -> Optional[OpennemDataSet]:
    engine = get_database_engine()

    query = country_stats_query(StatTypes.CPI)

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    stats = [
        DataQueryResult(interval=i[0], result=i[1], group_by=i[2] if len(i) > 1 else None)
        for i in row
    ]

    if len(stats) < 1:
        raise Exception("No results from query: {}".format(query))

    result = stats_factory(
        stats,
        code="au.cpi",
        network=NetworkNEM,
        interval=human_to_interval("1Q"),
        period=human_to_period("all"),
        units=get_unit("cpi"),
        group_field="gov",
    )

    return result


def power_flows_week(
    network: NetworkSchema,
    network_region_code: str,
    date_range: ScadaDateRange,
) -> Optional[OpennemDataSet]:
    engine = get_database_engine()

    query = interconnector_flow_power_query(
        network_region=network_region_code, date_range=date_range, network=network
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    stats: List[RegionFlowResult] = [
        RegionFlowResult(
            interval=i[0],
            generated=i[1],
            flow_from=i[2],
            flow_to=i[3] if len(i) > 1 else None,
        )
        for i in row
    ]

    if len(stats) < 1:
        raise Exception("No results from query: {}".format(query))

    stats_grouped = net_flows(network_region_code, stats)

    imports = stats_grouped["imports"]
    exports = stats_grouped["exports"]

    result = stats_factory(
        imports,
        # code=network_region_code or network.code,
        network=network,
        period=human_to_period("7d"),
        interval=human_to_interval("5m"),
        units=get_unit("power"),
        region=network_region_code,
        fueltech_group=True,
    )

    if not result:
        raise Exception("No results")

    result_exports = stats_factory(
        exports,
        # code=network_region_code or network.code,
        network=network,
        period=human_to_period("7d"),
        interval=human_to_interval("5m"),
        units=get_unit("power"),
        region=network_region_code,
        fueltech_group=True,
    )

    result.append_set(result_exports)

    return result


def power_week(
    network_code: str = "WEM",
    network_region_code: str = None,
    period: Optional[TimePeriod] = None,
    date_range: Optional[ScadaDateRange] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
) -> Optional[OpennemDataSet]:
    engine = get_database_engine()

    network = network_from_network_code(network_code)
    interval = human_to_interval("{}m".format(network.interval_size))
    units = get_unit("power")
    network_region = None

    query = power_network_fueltech_query(
        network=network,
        networks_query=networks_query,
        period=period,
        date_range=date_range,
        interval=interval,
        network_region=network_region_code,
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    stats = [
        DataQueryResult(interval=i[0], result=i[2], group_by=i[1] if len(i) > 1 else None)
        for i in row
    ]

    if len(stats) < 1:
        raise Exception("No results from query: {}".format(query))

    result = stats_factory(
        stats,
        # code=network_region_code or network.code,
        network=network,
        interval=interval,
        period=period,
        units=units,
        region=network_region,
        fueltech_group=True,
    )

    if not result:
        raise Exception("No results")

    # price

    query = price_network_query(
        network=network,
        networks_query=networks_query,
        period=period,
        date_range=date_range,
        interval=interval,
        network_region=network_region_code,
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    stats_price = [
        DataQueryResult(interval=i[0], result=i[2], group_by=i[1] if len(i) > 1 else None)
        for i in row
    ]

    stats_market_value = stats_factory(
        stats=stats_price,
        code=network_region_code or network.code.lower(),
        units=get_unit("price_energy_mega"),
        network=network,
        interval=interval,
        region=network_region_code,
        period=period,
    )

    result.append_set(stats_market_value)

    return result


def energy_fueltech_daily(
    network: NetworkSchema,
    year: Optional[int] = None,
    network_region_code: Optional[str] = None,
    interval_size: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
    date_range: Optional[ScadaDateRange] = None,
) -> Optional[OpennemDataSet]:
    engine = get_database_engine()
    period: TimePeriod = human_to_period("1Y")
    units = get_unit("energy_giga")

    interval = None

    if interval_size:
        interval = human_to_interval(interval_size)

    query = energy_network_fueltech_query(
        year=year,
        interval=interval,
        network=network,
        network_region=network_region_code,
        networks_query=networks_query,
        date_range=date_range,
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    results_energy = [
        DataQueryResult(interval=i[0], group_by=i[1], result=i[2] if len(i) > 1 else None)
        for i in row
    ]

    results_market_value = [
        DataQueryResult(interval=i[0], group_by=i[1], result=i[3] if len(i) > 1 else None)
        for i in row
    ]

    results_emissions = [
        DataQueryResult(interval=i[0], group_by=i[1], result=i[4] if len(i) > 1 else None)
        for i in row
    ]

    if len(results_energy) < 1:
        raise Exception("No results from query: {}".format(query))

    stats = stats_factory(
        stats=results_energy,
        units=units,
        network=network,
        fueltech_group=True,
        interval=interval,
        region=network_region_code,
        period=period,
        # code=network.code.lower(),
    )

    if not stats:
        return None

    stats_market_value = stats_factory(
        stats=results_market_value,
        units=get_unit("market_value"),
        network=network,
        fueltech_group=True,
        interval=interval,
        region=network_region_code,
        period=period,
        code=network.code.lower(),
    )

    stats.append_set(stats_market_value)

    stats_emissions = stats_factory(
        stats=results_emissions,
        units=get_unit("emissions"),
        network=network,
        fueltech_group=True,
        interval=interval,
        region=network_region_code,
        period=period,
        code=network.code.lower(),
    )

    stats.append_set(stats_emissions)

    return stats


def energy_interconnector_region_daily(
    network: NetworkSchema,
    network_region_code: str,
    year: Optional[int] = None,
    interval_size: Optional[str] = None,
    networks_query: Optional[List[NetworkSchema]] = None,
    date_range: Optional[ScadaDateRange] = None,
) -> Optional[OpennemDataSet]:
    engine = get_database_engine()
    period: TimePeriod = human_to_period("1Y")
    units = get_unit("energy_giga")

    interval = None

    if interval_size:
        interval = human_to_interval(interval_size)

    query = energy_network_fueltech_query(
        interconnector=True,
        year=year,
        interval=interval,
        network=network,
        network_region=network_region_code,
        networks_query=networks_query,
        date_range=date_range,
    )

    with engine.connect() as c:
        logger.debug(query)
        row = list(c.execute(query))

    stats: List[RegionFlowResult] = [
        RegionFlowResult(
            interval=i[0],
            generated=i[1],
            flow_from=i[2],
            flow_to=i[3] if len(i) > 1 else None,
        )
        for i in row
    ]

    if len(stats) < 1:
        raise Exception("No results from query: {}".format(query))

    stats_grouped = net_flows(network_region_code, stats)
    # stats_grouped = net_flows_energy(network_region_code, stats)

    imports = stats_grouped["imports"]
    exports = stats_grouped["exports"]

    result = stats_factory(
        imports,
        # code=network_region_code or network.code,
        network=network,
        period=period,
        interval=interval,
        units=units,
        region=network_region_code,
        fueltech_group=True,
    )

    # Bail early on no interconnector
    # don't error
    if not result:
        return result

    result_exports = stats_factory(
        exports,
        # code=network_region_code or network.code,
        network=network,
        period=period,
        interval=interval,
        units=units,
        region=network_region_code,
        fueltech_group=True,
    )

    result.append_set(result_exports)

    return result



from opennem.db import SessionLocal
from opennem.db.models.opennem import BomStation


def get_stations_priority():
    session = SessionLocal()

    stations = session.query(BomStation).filter(BomStation.priority < 2).all()

    return stations


def get_stations():
    session = SessionLocal()

    stations = session.query(BomStation).filter(BomStation.priority >= 2).all()

    return stations



# pylint: disable=no-member
"""Update energy sum functions

Revision ID: 4ee58f229572
Revises: 7c5a5ddfb9e1
Create Date: 2020-10-20 19:51:14.874586

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "4ee58f229572"
down_revision = "7c5a5ddfb9e1"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        drop function if exists agg_power(agg_state point, el numeric, interval_size_minutes int) cascade;
        drop function if exists agg_power(agg_state numeric[], el numeric, text) cascade;
        drop function if exists agg_power(agg_state numeric[], el numeric) cascade;

        create function agg_power(agg_state numeric[], el numeric)
        returns numeric[]
        immutable
        language plpgsql
        as $$
        begin
            agg_state := agg_state || el;
            return agg_state;
        end;
        $$;
    """
    )

    op.execute(
        """

        drop function if exists onenergy_finalfunc(agg_state point) cascade;

        drop function if exists onenergy_finalfunc(agg_state numeric[]);

        create function onenergy_finalfunc(agg_state numeric[])
        returns float8
        immutable
        strict
        language plpgsql
        as $$
        declare
            result_int numeric;
            divis numeric;
            bucket_size_minutes int := 0;
            m numeric;
            cur_index int := 1;
            current_sum numeric := 0;
            power_interval numeric;
            num_intervals int;
            a float := 0;
            b float := 1;
            n float := 10;
            w int := 0;
            h float := (b - a) / n;
        begin
            bucket_size_minutes := extract(epoch FROM '1 day'::interval) / 60;

            num_intervals := array_length(agg_state, 1);

            FOREACH m IN ARRAY agg_state
            LOOP
                IF cur_index = 1 then
                    w := 1;
                elsif cur_index = (n + 1) then
                    w := 1;
                elsif (cur_index % 2) = 1 then
                    w := 2;
                else
                    w := 4;
                END IF;

                cur_index := cur_index + 1;

                divis := (60 / (bucket_size_minutes / num_intervals));

                if m > 0 and divis > 0 then
                    power_interval = m / divis;
                else
                    power_interval = 0;
                end if;

                current_sum = current_sum + (w * power_interval);
            END LOOP;

            result_int := h * current_sum / 6 * 10;

            return round(result_int, 2);
        end;
        $$;
    """
    )

    op.execute(
        """
        create aggregate on_energy_sum (numeric)
        (
            sfunc = agg_power,
            stype = numeric[],
            finalfunc = onenergy_finalfunc,
            initcond = '{}'
        );
    """
    )


def downgrade():
    op.execute(
        "drop aggregate if exists on_energy_sum (numeric, text) cascade"
    )
    op.execute(
        "drop function if exists agg_power (agg_state numeric[], el numeric) cascade"
    )
    op.execute(
        "drop function if exists onenergy_finalfunc (agg_state numeric[]) cascade"
    )



import io
import logging
from io import BytesIO, StringIO
from os import makedirs
from pathlib import Path
from typing import Optional, Union

from opennem.settings import settings
from opennem.utils.mime import decode_bytes

logger = logging.getLogger(__name__)


def write_to_local(
    file_path: str, data: Union[StringIO, bytes, BytesIO, str]
) -> int:
    save_folder = settings.static_folder_path

    save_file_path = Path(save_folder) / file_path.lstrip("/")

    dir_path = save_file_path.resolve().parent

    if not dir_path.is_dir():
        makedirs(dir_path)

    write_data: Optional[str] = None

    if isinstance(data, StringIO):
        write_data = data.getvalue()
    elif isinstance(data, BytesIO):
        write_data = decode_bytes(data.getvalue())
    elif isinstance(data, bytes):
        write_data = decode_bytes(data)
    elif data:
        write_data = data

    bytes_written = 0

    if not write_data:
        logger.info("No data to write to {}".format(file_path))
        return 0

    with open(save_file_path, "w") as fh:
        bytes_written += fh.write(write_data)

    logger.info("Wrote {} to {}".format(bytes_written, save_file_path))

    return bytes_written



"""

NEMWEB Data ingress into OpenNEM format


"""

import logging
from datetime import datetime
from itertools import groupby
from typing import Any, Dict, List, Optional

from scrapy import Spider
from sqlalchemy.dialects.postgresql import insert

from opennem.core.networks import NetworkNEM
from opennem.core.normalizers import clean_float, normalize_duid
from opennem.db import SessionLocal, get_database_engine
from opennem.db.models.opennem import BalancingSummary, Facility, FacilityScada
from opennem.importer.rooftop import rooftop_remap_regionids
from opennem.schema.network import NetworkSchema
from opennem.utils.dates import parse_date
from opennem.utils.numbers import float_to_str
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


# Helpers


def unit_scada_generate_facility_scada(
    records,
    spider=None,
    network: NetworkSchema = NetworkNEM,
    interval_field: str = "SETTLEMENTDATE",
    facility_code_field: str = "DUID",
    date_format: Optional[str] = None,
    power_field: Optional[str] = None,
    energy_field: Optional[str] = None,
    is_forecast: bool = False,
    primary_key_track: bool = False,
    groupby_filter: bool = True,
    created_by: str = None,
    limit: int = 0,
    duid: str = None,
) -> List[Dict]:
    created_at = datetime.now()
    primary_keys = []
    return_records = []

    created_by = ""

    if spider and hasattr(spider, "name"):
        created_by = spider.name

    for row in records:

        trading_interval = parse_date(
            row[interval_field],
            network=network,
            dayfirst=False,
            date_format=date_format,
        )

        # if facility_code_field not in row:
        # logger.error("Invalid row no facility_code")
        # continue

        facility_code = normalize_duid(row[facility_code_field])

        if duid and facility_code != duid:
            continue

        if primary_key_track:
            pkey = (trading_interval, facility_code)

            if pkey in primary_keys:
                continue

            primary_keys.append(pkey)

        generated = None

        if power_field and power_field in row:
            generated = clean_float(row[power_field])

            if generated:
                generated = float_to_str(generated)

        energy = None

        if energy_field and energy_field in row:
            energy = clean_float(row[energy_field])

            if energy:
                energy = float_to_str(energy)

        __rec = {
            "created_by": created_by,
            "created_at": created_at,
            "updated_at": None,
            "network_id": network.code,
            "trading_interval": trading_interval,
            "facility_code": facility_code,
            "generated": generated,
            "eoi_quantity": energy,
            "is_forecast": is_forecast,
        }

        return_records.append(__rec)

        if limit > 0 and len(return_records) >= limit:
            break

    if not groupby_filter:
        return return_records

    return_records_grouped = {}

    for pk_values, rec_value in groupby(
        return_records,
        key=lambda r: (
            r.get("network_id"),
            r.get("trading_interval"),
            r.get("facility_code"),
        ),
    ):
        if pk_values not in return_records_grouped:
            return_records_grouped[pk_values] = list(rec_value).pop()

    return_records = list(return_records_grouped.values())

    return return_records


def generate_balancing_summary(
    records: List[Dict],
    spider: Spider,
    interval_field: str = "SETTLEMENTDATE",
    network_region_field: str = "REGIONID",
    price_field: Optional[str] = None,
    network: NetworkSchema = NetworkNEM,
    limit: int = 0,
) -> List[Dict]:
    created_at = datetime.now()
    # primary_keys = []
    return_records = []

    created_by = ""

    if spider and hasattr(spider, "name"):
        created_by = spider.name

    for row in records:

        trading_interval = parse_date(row[interval_field], network=network, dayfirst=False)

        network_region = None

        if network_region_field and network_region_field in row:
            network_region = row[network_region_field]

        price = None

        if price_field and price_field in row:
            price = clean_float(row[price_field])

            if price:
                price = float_to_str(price)

        __rec = {
            "created_by": created_by,
            "created_at": created_at,
            "updated_at": None,
            "network_id": network.code,
            "network_region": network_region,
            "trading_interval": trading_interval,
            "price": price,
        }

        return_records.append(__rec)

        if limit > 0 and len(return_records) >= limit:
            break

    return return_records


# Processors


def get_interconnector_facility(facility_code: str) -> Facility:
    session = SessionLocal()

    _fac = session.query(Facility).filter_by(code=facility_code).one_or_none()

    if not _fac:
        raise Exception("Could not find facility {}".format(facility_code))

    return _fac


def process_dispatch_interconnectorres(table: Dict, spider: Spider) -> Dict:
    session = SessionLocal()
    engine = get_database_engine()

    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]

    records_to_store = []

    for record in records:
        ti_value = None

        if "SETTLEMENTDATE" in record:
            ti_value = record["SETTLEMENTDATE"]

        if "RUN_DATETIME" in record:
            ti_value = record["RUN_DATETIME"]

        if not ti_value:
            raise Exception("Require a trading interval")

        trading_interval = parse_date(ti_value, network=NetworkNEM, dayfirst=False)

        if not trading_interval:
            continue

        facility_code = normalize_duid(record["INTERCONNECTORID"])
        power_value = clean_float(record["METEREDMWFLOW"])

        records_to_store.append(
            {
                "network_id": "NEM",
                "created_by": spider.name,
                "facility_code": facility_code,
                "trading_interval": trading_interval,
                "generated": power_value,
            }
        )

    # remove duplicates
    return_records_grouped = {}

    for pk_values, rec_value in groupby(
        records_to_store,
        key=lambda r: (
            r.get("trading_interval"),
            r.get("network_id"),
            r.get("facility_code"),
        ),
    ):
        if pk_values not in return_records_grouped:
            return_records_grouped[pk_values] = list(rec_value).pop()

    records_to_store = list(return_records_grouped.values())

    # insert
    stmt = insert(FacilityScada).values(records_to_store)
    stmt.bind = engine
    stmt = stmt.on_conflict_do_update(
        index_elements=["trading_interval", "network_id", "facility_code"],
        set_={"generated": stmt.excluded.generated},
    )

    try:
        session.execute(stmt)
        session.commit()
    except Exception as e:
        logger.error("Error inserting records")
        logger.error(e)
        return {"num_records": 0}
    finally:
        session.close()

    return {"num_records": len(records_to_store)}


def process_case_solutions(table: Dict, spider: Spider):
    pass


def process_pre_ap_price(table: Dict, spider: Spider) -> int:
    session = SessionLocal()
    engine = get_database_engine()

    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]

    limit = None
    records_to_store = []
    records_processed = 0
    primary_keys = []

    for record in records:
        trading_interval = parse_date(
            record["SETTLEMENTDATE"],
            network=NetworkNEM,
            dayfirst=False,
            date_format="%Y/%m/%d %H:%M:%S",
        )

        if not trading_interval:
            continue

        _pk = set([trading_interval, record["REGIONID"]])

        if _pk in primary_keys:
            continue

        primary_keys.append(_pk)

        records_to_store.append(
            {
                "network_id": "NEM",
                "created_by": spider.name,
                "network_region": record["REGIONID"],
                "trading_interval": trading_interval,
                "price": record["PRE_AP_ENERGY_PRICE"],
            }
        )

        records_processed += 1

        if limit and records_processed >= limit:
            logger.info("Reached limit of: {} {}".format(limit, records_processed))
            break

    stmt = insert(BalancingSummary).values(records_to_store)
    stmt.bind = engine
    stmt = stmt.on_conflict_do_update(
        index_elements=["trading_interval", "network_id", "network_region"],
        set_={"price": stmt.excluded.price},
    )

    try:
        session.execute(stmt)
        session.commit()
    except Exception as e:
        logger.error("Error inserting records")
        logger.error(e)
        return 0
    finally:
        session.close()

    return len(records_to_store)


def process_unit_scada(table: Dict[str, Any], spider: Spider) -> Dict:
    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]
    item = dict()

    item["table_schema"] = FacilityScada
    item["update_fields"] = ["generated"]
    item["records"] = unit_scada_generate_facility_scada(
        records,
        spider,
        power_field="SCADAVALUE",
        network=NetworkNEM,
        date_format="%Y/%m/%d %H:%M:%S",
    )
    item["content"] = ""

    return item


def process_unit_solution(table: Dict[str, Any], spider: Spider) -> Dict:
    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]
    item = dict()

    item["table_schema"] = FacilityScada
    item["update_fields"] = ["generated"]
    item["records"] = unit_scada_generate_facility_scada(
        records,
        spider,
        network=NetworkNEM,
        interval_field="SETTLEMENTDATE",
        facility_code_field="DUID",
        power_field="INITIALMW",
    )
    item["content"] = None

    return item


def process_meter_data_gen_duid(table: Dict[str, Any], spider: Spider) -> Dict:
    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]
    item = dict()

    item["table_schema"] = FacilityScada
    item["update_fields"] = ["generated"]
    item["records"] = unit_scada_generate_facility_scada(
        records,
        spider,
        network=NetworkNEM,
        interval_field="INTERVAL_DATETIME",
        facility_code_field="DUID",
        power_field="MWH_READING",
    )
    item["content"] = None

    return item


def process_rooftop_actual(table: Dict[str, Any], spider: Spider) -> Dict:
    if "records" not in table:
        raise Exception("Invalid table no records")

    records = table["records"]
    item = dict()

    # Filter out for only measurements
    # as opposed to TYPE=SATELLITE
    records_filtered = list(filter(lambda x: x["TYPE"] == "MEASUREMENT", records))

    scada_records = unit_scada_generate_facility_scada(
        records_filtered,
        spider,
        network=NetworkNEM,
        interval_field="INTERVAL_DATETIME",
        facility_code_field="REGIONID",
        power_field="POWER",
        # don't filter down on duid since they'll conflict
        # we need to sum aggs
        groupby_filter=False,
    )

    scada_records = [rooftop_remap_regionids(i) for i in scada_records]
    return_records_grouped = {}

    # Merge and group
    for pk_values, rec_value in groupby(
        scada_records,
        key=lambda r: (
            r.get("facility_code"),
            r.get("network_id"),
            r.get("trading_interval"),
        ),
    ):
        if pk_values not in return_records_grouped:
            rec_values = list(rec_value)
            return_records_grouped[pk_values] = rec_values.pop()
            return_records_grouped[pk_values]["generated"] = 0.0

            for _rec in rec_values:
                _new_val = clean_float(_rec["generated"])

                if _new_val:
                    return_records_grouped[pk_values]["generated"] += _new_val

    return_records = list(return_records_grouped.values())

    item["table_schema"] = FacilityScada
    item["update_fields"] = ["generated"]
    item["records"] = return_records
    item["content"] = None

    return item


TABLE_PROCESSOR_MAP = {
    "DISPATCH_INTERCONNECTORRES": "process_dispatch_interconnectorres",
    # "P5MIN_INTERCONNECTORSOLN": "process_dispatch_interconnectorres",
    "METER_DATA_GEN_DUID": "process_meter_data_gen_duid",
    "DISPATCH_CASE_SOLUTION": "process_case_solutions",
    "DISPATCH_UNIT_SCADA": "process_unit_scada",
    "DISPATCH_UNIT_SOLUTION": "process_unit_solution",
    # "DISPATCH_PRE_AP_PRICE": "process_pre_ap_price",
    "ROOFTOP_ACTUAL": "process_rooftop_actual",
    "DISPATCH_PRICE": "process_pre_ap_price",
}


class NemwebUnitScadaOpenNEMStorePipeline(object):
    @check_spider_pipeline
    def process_item(self, item, spider=None):
        if not item:
            msg = "NemwebUnitScadaOpenNEMStorePipeline"
            if spider and hasattr(spider, "name"):
                msg = spider.name
            logger.error("No item in pipeline: {}".format(msg))
            return {}

        if "tables" not in item:
            print(item)
            raise Exception("Invalid item - no tables located")

        if not isinstance(item["tables"], dict):
            raise Exception("Invalid item - no tables located")

        tables = item["tables"]

        ret = []

        for table in tables.values():
            if "name" not in table:
                logger.info("Invalid table found")
                continue

            table_name = table["name"]

            if table_name not in TABLE_PROCESSOR_MAP:
                logger.info("No processor for table %s", table_name)
                continue

            process_meth = TABLE_PROCESSOR_MAP[table_name]

            if process_meth not in globals():
                logger.info("Invalid processing function %s", process_meth)
                continue

            logger.info("processing table {}".format(table_name))

            record_item = None

            try:
                record_item = globals()[process_meth](table, spider=spider)
            except Exception as e:
                logger.error(e)

            if record_item:
                ret.append(record_item)

        return ret



# pylint: disable=no-member
"""
Interconnector metadata on facility

Revision ID: 000e7fdab7e2
Revises: 8746bdb27550
Create Date: 2020-12-24 15:01:38.583540

"""
import geoalchemy2
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "000e7fdab7e2"
down_revision = "8746bdb27550"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        "facility", sa.Column("interconnector", sa.Boolean(), nullable=True)
    )
    op.add_column(
        "facility",
        sa.Column("interconnector_region_to", sa.Text(), nullable=True),
    )
    op.create_index(
        op.f("ix_facility_interconnector"),
        "facility",
        ["interconnector"],
        unique=False,
    )
    op.create_index(
        op.f("ix_facility_interconnector_region_to"),
        "facility",
        ["interconnector_region_to"],
        unique=False,
    )


def downgrade() -> None:
    op.drop_index(
        op.f("ix_facility_interconnector_region_to"), table_name="facility"
    )
    op.drop_index(op.f("ix_facility_interconnector"), table_name="facility")
    op.drop_column("facility", "interconnector_region_to")
    op.drop_column("facility", "interconnector")



import csv
import logging

from sqlalchemy.dialects.postgresql import insert

from opennem.core.normalizers import clean_float
from opennem.db import SessionLocal, get_database_engine
from opennem.db.models.opennem import BalancingSummary
from opennem.schema.network import NetworkWEM
from opennem.utils.dates import parse_date
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


class WemStorePulse(object):
    @check_spider_pipeline
    def process_item(self, item, spider=None):

        s = SessionLocal()

        csvreader = csv.DictReader(item["content"].split("\n"))

        records_to_store = []
        primary_keys = []

        for row in csvreader:
            trading_interval = parse_date(
                row["TRADING_DAY_INTERVAL"], network=NetworkWEM, dayfirst=False
            )

            if trading_interval not in primary_keys:
                forecast_load = clean_float(row["FORECAST_EOI_MW"])

                records_to_store.append(
                    {
                        "created_by": spider.name,
                        "trading_interval": trading_interval,
                        "network_id": "WEM",
                        "network_region": "WEM",
                        "forecast_load": forecast_load,
                        # generation_scheduled=row["Scheduled Generation (MW)"],
                        # generation_total=row["Total Generation (MW)"],
                        "price": clean_float(row["PRICE"]),
                    }
                )
                primary_keys.append(trading_interval)

        stmt = insert(BalancingSummary).values(records_to_store)
        stmt.bind = get_database_engine()
        stmt = stmt.on_conflict_do_update(
            index_elements=[
                "trading_interval",
                "network_id",
                "network_region",
            ],
            set_={
                "price": stmt.excluded.price,
                "forecast_load": stmt.excluded.forecast_load,
            },
        )

        try:
            s.execute(stmt)
            s.commit()
        except Exception as e:
            logger.error("Error inserting records")
            logger.error(e)
            return 0
        finally:
            s.close()

        return len(records_to_store)
